@@ -1,4 +1,4 @@
-/* Copyright (c) 1998, 1999, 2000 Thorsten Kukuk
+/* Copyright (c) 1998, 1999, 2000, 2001 Thorsten Kukuk
    This file is part of ypbind-mt.
    Author: Thorsten Kukuk <kukuk@suse.de>
 
@@ -40,6 +40,7 @@
 #include <sys/uio.h>
 #include <sys/stat.h>
 #include <sys/ioctl.h>
+#include <sys/socket.h>
 #if defined(HAVE_SYS_FILIO_H)
 #include <sys/filio.h>
 #endif
@@ -88,6 +89,7 @@ static pthread_mutex_t search_lock = PTHREAD_MUTEX_INITIALIZER;
 static void do_broadcast (struct binding *list);
 static void ping_all (struct binding *list);
 
+/* this is called from the RPC thread (ypset). */
 void
 change_binding (const char *domain, ypbind_binding *binding)
 {
@@ -192,7 +194,7 @@ find_domain (const char *domain, ypbind_resp *result)
 		  break;
 		}
 	      /* Look, if we could find a new server for this domain.
-		 But only, if the other thread is searching already */
+		 But only, if the other thread is not searching already */
 	      pthread_rdwr_runlock_np (&domainlock);
 	      if (pthread_mutex_trylock (&search_lock) == 0)
 		{
@@ -204,6 +206,8 @@ find_domain (const char *domain, ypbind_resp *result)
 		    ping_all (&domainlist[i]);
 		  pthread_mutex_unlock (&search_lock);
 		  ++second;
+		  /* Get the read lock again for next run. */
+		  pthread_rdwr_rlock_np (&domainlock);
 		  continue;
 		}
 	      else
@@ -220,7 +224,7 @@ find_domain (const char *domain, ypbind_resp *result)
 }
 
 static void
-close_bindingfile (const char *domain)
+remove_bindingfile (const char *domain)
 {
   char path[strlen (BINDINGDIR) + strlen (domain) + 10];
 
@@ -240,21 +244,11 @@ update_bindingfile (struct binding *entry)
   struct iovec iov[2];
   struct ypbind_resp ybres;
   char path1[MAXPATHLEN + 1], path2[MAXPATHLEN + 1];
-  int fd1, fd2, len;
+  int fd, len;
 
   sprintf (path1, "%s/%s.1", BINDINGDIR, entry->domain);
   sprintf (path2, "%s/%s.2", BINDINGDIR, entry->domain);
 
-  if ((fd1 = open(path1, O_CREAT | O_RDWR | O_TRUNC, FILE_MODE )) == -1)
-    return;
-
-  if ((fd2 = open(path2, O_CREAT | O_RDWR | O_TRUNC, FILE_MODE )) == -1)
-    {
-      close (fd1);
-      unlink (path1);
-      return;
-    }
-
   iov[0].iov_base = (caddr_t) &sport;
   iov[0].iov_len = sizeof (sport);
   iov[1].iov_base = (caddr_t) &ybres;
@@ -269,28 +263,48 @@ update_bindingfile (struct binding *entry)
       memcpy (&ybres.ypbind_resp_u.ypbind_bindinfo.ypbind_binding_port,
 	      &entry->server[entry->active].port, sizeof (unsigned short int));
     }
-  else
+  else if (entry->active == -2) /* ypset was used */
     {
       memcpy (&ybres.ypbind_resp_u.ypbind_bindinfo.ypbind_binding_addr,
 	      &entry->ypset.addr, sizeof (struct in_addr));
       memcpy (&ybres.ypbind_resp_u.ypbind_bindinfo.ypbind_binding_port,
 	      &entry->ypset.port, sizeof (unsigned short int));
     }
+  else
+    {
+       /* This should not happen. Remove binding files which means,
+          libc will query ypbind direct. */
+       unlink (path1);
+       unlink (path2);
+       log_msg (LOG_ERR, "INTERNAL ERROR: update_bindingfile called without valid data!");
+       return;
+    }
 
   len = iov[0].iov_len + iov[1].iov_len;
-  if (writev (fd1, iov, 2) != len )
-    {
-      log_msg (LOG_ERR, "writev (fd1): %s", strerror (errno));
-      unlink (path1);
-    }
-  close (fd1);
 
-  if (writev (fd2, iov, 2) != len )
-    {
-      log_msg (LOG_ERR, "writev (fd2): %s", strerror (errno));
-      unlink (path2);
+  if ((fd = open(path1, O_CREAT | O_RDWR | O_TRUNC, FILE_MODE )) != -1)
+    {  
+      if (writev (fd, iov, 2) != len )
+        {
+          log_msg (LOG_ERR, "writev (%s): %s", path1, strerror (errno));
+          unlink (path1);
+        }
+      close (fd);
     }
-  close (fd2);
+  else
+    log_msg (LOG_ERR, "open(%s): %s", path1, strerror (errno));
+
+  if ((fd = open(path2, O_CREAT | O_RDWR | O_TRUNC, FILE_MODE )) != -1)
+    {
+      if (writev (fd, iov, 2) != len )
+        {
+          log_msg (LOG_ERR, "writev (%s): %s", path2, strerror (errno));
+          unlink (path2);
+        }
+      close (fd);
+    }
+  else
+    log_msg (LOG_ERR, "open(%s): %s", path2, strerror (errno));
 }
 
 void
@@ -583,7 +597,7 @@ do_broadcast (struct binding *list)
 
   if (status != RPC_SUCCESS)
     {
-      close_bindingfile(list->domain);
+      remove_bindingfile(list->domain);
       log_msg (LOG_ERR, "broadcast: %s.", clnt_sperrno(status));
     }
   else
@@ -800,7 +814,7 @@ ping_all (struct binding *list)
   free (pings);
 
   if (!found)
-    close_bindingfile(list->domain);
+    remove_bindingfile(list->domain);
 
   return;
 }
@@ -890,7 +904,7 @@ ping_all (struct binding *list)
       ++i;
       if (i == _MAXSERVER)
         {
-          close_bindingfile(list->domain);
+          remove_bindingfile(list->domain);
           return;
         }
     }
@@ -959,8 +973,11 @@ test_bindings (void *param __attribute__ ((unused)))
 
 	  if (domainlist[i].active != -1)
 	    {
-	      if (lastcheck)
+	      /* The binding is in use, check if it is still valid and
+	         the fastest one. */
+	      if (lastcheck != 0)
 		{
+		  /* Check only if the current binding is still valid. */
 		  struct timeval time_out;
 
 		  time_out.tv_sec = 3;
@@ -971,10 +988,18 @@ test_bindings (void *param __attribute__ ((unused)))
 			      (caddr_t) &domain, (xdrproc_t) xdr_bool,
 			      (caddr_t) &out, time_out);
 		}
-	      if (lastcheck == 0 || status != RPC_SUCCESS || out != TRUE)
+	      /* time to search a new fastest server, but only if the current one was
+		 not set with ypset. We search in every case if the above check fails
+	         and the current data is not longer valid. */
+	      if ((lastcheck == 0 && domainlist[i].active != -2)
+		  || status != RPC_SUCCESS || out != TRUE)
 		{
-		  if (debug_flag && lastcheck)
+		  /* The current binding is not valid or it is time to search
+		     for a new, fast server. */
+		  if (debug_flag && lastcheck != 0)
 		    {
+		      /* Current active binding is not longer valid, print
+		         the old binding for debugging. */
 		      if (domainlist[i].use_broadcast)
 			log_msg (LOG_DEBUG,
 				 _("Server for domain '%s' doesn't answer."),
@@ -1000,16 +1025,19 @@ test_bindings (void *param __attribute__ ((unused)))
 		     changes :-( */
 		  pthread_rdwr_runlock_np (&domainlock);
 		  pthread_rdwr_wlock_np (&domainlock);
+		  /* We can destroy the client_handle since we are the
+		     only thread who uses it. */
 		  clnt_destroy (domainlist[i].client_handle);
 		  domainlist[i].client_handle = NULL;
 		  if (domainlist[i].active == -2)
 		    {
+		      /* We can give this free, server does not answer any
+			 longer. */
+		      domainlist[i].active = -1;
 		      if (domainlist[i].ypset.host != NULL)
 			free (domainlist[i].ypset.host);
 		      domainlist[i].ypset.host = NULL;
 		    }
-		  domainlist[i].active = -1;
-		  close_bindingfile (domain);
 		  /* And give the write lock away, search a new host and get
 		     the read lock again */
 		  pthread_rdwr_wunlock_np (&domainlock);
@@ -1024,6 +1052,8 @@ test_bindings (void *param __attribute__ ((unused)))
 	    }
 	  else
 	    {
+	      /* there is no binding for this domain, try to find a new
+		 server */
 	      pthread_rdwr_runlock_np (&domainlock);
 	      pthread_mutex_lock (&search_lock);
 	      if (domainlist[i].use_broadcast)
@@ -1033,7 +1063,9 @@ test_bindings (void *param __attribute__ ((unused)))
 	      pthread_mutex_unlock (&search_lock);
 	      pthread_rdwr_rlock_np (&domainlock);
 	    }
-	}
+	} /* end for () all domains */
+
       pthread_rdwr_runlock_np (&domainlock);
-    }
+
+    } /* end while() endless loop */
 }
